from __future__ import unicode_literals

from django.conf.urls import url
from django.core.paginator import Paginator, EmptyPage
from django.utils.decorators import method_decorator
from django.views.decorators.http import require_http_methods
from django.views.decorators.csrf import ensure_csrf_cookie, csrf_exempt

from cgi import parse_header
from collections import defaultdict
try:
    import ujson as json
except ImportError:
    import json

from . import http

def accepts(*verbs):
    '''Annotate a method with the HTTP verbs it accepts, and enforce it.'''
    def _inner(method):
        setattr(method, '_accepts', verbs)
        return method_decorator(
            require_http_methods([x.upper() for x in verbs])
        )(method)
    return _inner

class BasePublisher(object):
    CSRF = True

    ACTION_PATTERN = r'\w+'
    OBJECT_PATTERN = r'[-\w]+'
    ARGUMENT_PATTERN = r'.+?'

    def __init__(self, request, *args, **kwargs):
        self.request = request
        self.args = args
        self.kwargs = kwargs

    @classmethod
    def build_view(cls):
        '''Builds the view function for this publisher.'''
        def view(request, *args, **kwargs):
            '''A wrapper view to instantiate and dispatch'''
            self = cls(request, *args, **kwargs)
            return self.dispatch(request, **kwargs)

        if cls.CSRF:
            view = ensure_csrf_cookie(view)
        else:
            view = csrf_exempt(view)
        return view

    @classmethod
    def patterns(cls, api_name=None):
        '''
        Add this to your url patterns like:
            ( '^foo/', include(mypublisher.patterns()), ),
        /                       default object list
        /(action)/              list operation
        /(action)/(option)/     list operation with extra argument
        /object/(id)/           instance view
        /object/(id)/(action)/  custom action on instance
        '''
        view = cls.build_view()

        name = getattr(cls, 'api_name', cls.__name__.lower())
        if api_name:
            name = '%s_%s' % (api_name, name)

        return [
            url(r'^object/(?P<object_id>%s)/(?P<action>%s)/(?P<argument>%s)/?$' % (cls.OBJECT_PATTERN, cls.ACTION_PATTERN, cls.ARGUMENT_PATTERN),
                view,
                name='%s_object_action_arg' % name
            ),
            url(r'^object/(?P<object_id>%s)/(?P<action>%s)/?$' % (cls.OBJECT_PATTERN, cls.ACTION_PATTERN),
                view,
                name='%s_object_action' % name
            ),
            url(r'^object/(?P<object_id>%s)/?$' % (cls.OBJECT_PATTERN,),
                view,
                name='%s_object_default' % name
            ),
            url(r'^(?P<action>%s)/(?P<argument>%s)/?$' % (cls.ACTION_PATTERN, cls.ARGUMENT_PATTERN),
                view,
                name='%s_list_action_arg' % name
            ),
            url(r'^(?P<action>%s)/?$' % (cls.ACTION_PATTERN,),
                view,
                name='%s_list_action' % name
            ),
            url(r'^$',
                view,
                name='%s_list_default' % name
            ),
        ]


    def dispatch(self, request, action='default', object_id=None, **kwargs):
        '''View dispatcher called by Django'''
        self.action = action
        self.object_id = object_id
        method = request.method.lower()
        self.mode = prefix = 'object' if object_id else 'list'
        handler = getattr(self, '%s_%s_%s' % (prefix, method, action), None)
        if handler is None:
            # See if there's a method agnostic handler
            handler = getattr(self, '%s_%s' % (prefix, action), None)
        if handler is None:
            return http.NotFound()
        # Do we need to pass any of this?
        return self.execute(handler)

    def execute(self, handler):
        '''This allows wrapping calls to handler functions'''
        try:
            return handler(self.request,
                action=self.action,
                object_id=self.object_id,
            )
        except http.BaseHttpResponse as response:
            return response

    @classmethod
    def index(cls):
        '''Return details about which handlers exist on this publisher.'''
        list_handlers = defaultdict(list)
        object_handlers = defaultdict(list)
        for name in dir(cls):
            fnc = getattr(cls, name)
            if not callable(fnc):
                continue
            parts = name.split('_')

            if parts[0] == 'list':
                if len(parts) == 2:
                    list_handlers[parts[1]].extend(getattr(fnc, '_accepts', ['ALL']))
                else:
                    list_handlers[parts[2]].append(parts[1])

            elif parts[0] == 'object':
                if len(parts) == 2:
                    object_handlers[parts[1]].extend(getattr(fnc, '_accepts', ['ALL']))
                else:
                    object_handlers[parts[2]].append(parts[1])

        return {
            'list': list_handlers,
            'object': object_handlers,
        }


class SimplePatternsMixin(object):
    '''
    A "flatter" set of url patterns for when your object IDs are numbers only.
    '''

    @classmethod
    def patterns(cls, api_name=None):
        view = cls.build_view()

        name = getattr(cls, 'api_name', cls.__name__.lower())
        if api_name:
            name = '%s_%s' % (api_name, name)

        return [
            url(r'^(?P<object_id>\d+)/(?P<action>\w+)/(?P<argument>.+?)/?$',
                view,
                name='%s_object_action_arg' % name
            ),
            url(r'^(?P<object_id>\d+)/(?P<action>\w+)/?$',
                view,
                name='%s_object_action' % name
            ),
            url(r'^(?P<object_id>\d+)/?$',
                view,
                name='%s_object_default' % name
            ),
            url(r'^(?P<action>\w+)/(?P<argument>.+?)/?$',
                view,
                name='%s_list_action_arg' % name
            ),
            url(r'^(?P<action>\w+)/?$',
                view,
                name='%s_list_action' % name
            ),
            url(r'^$',
                view,
                name='%s_list_default' % name
            ),
        ]


class Publisher(BasePublisher):
    '''Default API-style publisher'''
    LIMIT_PARAM = 'limit'
    OFFSET_PARAM = 'offset'
    PAGE_PARAM = 'page'

    response_class = http.HttpResponse

    # De/Serialising
    # Which content types will we attempt to parse?
    # The first in the list will be used for responses.
    CONTENT_TYPES = ['application/json', 'text/json']

    def dumps(self, data):
        '''How to parse content that matches our content types list.'''
        return json.dumps(data)
    def loads(self, data):
        '''Serialise data for responses.'''
        return json.loads(data)

    # Hooks for controlling which serialiser to use

    def get_serialiser(self):
        '''Return the serialiser instance to use for this request'''
        return self.serialiser

    def get_serialiser_kwargs(self):
        '''Allow passing of extra kwargs to serialiser calls'''
        return {
            'publisher': self,
        }

    # Object access

    def get_object_list(self): # pragma: no cover
        '''Return the object list appropriate for this request'''
        raise NotImplementedError

    def get_object(self, object_id): # pragma: no cover
        '''Return the object for the given id'''
        raise NotImplementedError

    # List filtering and sorting

    def filter_object_list(self, object_list):
        '''Hook to allow custom filtering of object lists'''
        return object_list

    def sort_object_list(self, object_list):
        '''Hook to allow custom sorting of object lists'''
        return object_list

    # Pagination
    def get_page(self, object_list):
        '''Return a paginated object list, along with some metadata'''
        page_size = getattr(self, 'page_size', None)
        if page_size is None:
            return {
                'meta': {},
                'objects': object_list,
            }
        max_page_size = getattr(self, 'max_page_size', page_size)
        page_size = int(self.request.GET.get(self.LIMIT_PARAM, page_size))
        page_size = max(0, min(page_size, max_page_size))
        page_num = 0
        try:
            page_num = int(self.request.GET[self.PAGE_PARAM])
        except ValueError:
            # Bad page - default to 0
            pass
        except KeyError:
            try:
                offset = int(self.request.GET[self.OFFSET_PARAM])
            except ValueError:
                # Bad page - default to 0
                pass
            except KeyError:
                # No value - default to 0
                pass
            else:
                page_num = offset // page_size

        paginator = Paginator(object_list, page_size, allow_empty_first_page=True)
        try:
            page = paginator.page(page_num + 1)
        except EmptyPage:
            raise http.NotFound()
        return {
            'meta': {
                'offset': page.start_index() - 1,
                'page': page_num,
                'total_pages': paginator.num_pages,
                'limit': page_size,
                'count': paginator.count,
                'has_next': page.has_next(),
                'has_prev': page.has_previous(),
            },
            'objects': page.object_list,
        }

    # Get the parsed request data

    def get_request_data(self, default=None):
        '''Retrieve data from request'''
        content_type, _ = parse_header(self.request.META.get('CONTENT_TYPE', ''))
        if content_type in self.CONTENT_TYPES:
            if not self.request.body:
                return default
            return self.loads(self.request.body)
        if self.request.method == 'GET':
            return self.request.GET
        return self.request.POST

    # Response helpers

    def create_response(self, content, **response_kwargs):
        '''Return a response, serialising the content'''
        response_class = response_kwargs.pop('response_class', self.response_class)
        response_kwargs.setdefault('content_type', self.CONTENT_TYPES[0])
        return response_class(self.dumps(content), **response_kwargs)

    def render_single_object(self, obj, serialiser=None, **response_kwargs):
        '''Helper to return a single object instance serialised.'''
        if serialiser is None:
            serialiser = self.get_serialiser()
        serialiser_kwargs = response_kwargs.pop('serialiser_kwargs', None)
        if serialiser_kwargs is None:
            serialiser_kwargs = self.get_serialiser_kwargs()
        data = serialiser.object_deflate(obj, **serialiser_kwargs)
        return self.create_response(data, **response_kwargs)

    # Default handlers

    def list_get_default(self, request, **kwargs):
        object_list = self.get_object_list()
        object_list = self.filter_object_list(object_list)
        object_list = self.sort_object_list(object_list)

        data = self.get_page(object_list)

        serialiser = self.get_serialiser()
        serialiser_kwargs = self.get_serialiser_kwargs()
        data['objects'] = serialiser.list_deflate(data['objects'], **serialiser_kwargs)
        return self.create_response(data)

    def object_get_default(self, request, object_id, **kwargs):
        '''Default object GET handler -- get object'''
        obj = self.get_object(object_id)
        return self.render_single_object(obj)




from django.conf.urls import url

from nap.publisher import Publisher
from nap import http

class RpcPublisher(Publisher):
    '''
    RPC Style Publisher.

    Single URL endpoint, determines handler based on X-RPC-Action header and
    HTTP method.
    '''

    @classmethod
    def patterns(cls, api_name=None):
        view = cls.build_view()

        if api_name:
            name = '%s_%s' % (api_name, cls.api_name)
        else:
            name = cls.api_name

        return [
            url(r'^$', view, name='%s_endpoint' % name),
        ]

    def dispatch(self, request):
        self.action = request.HTTP['X-RPC-ACTION']
        try:
            handler = getattr(self, 'do_%s' % self.action, None)
        except AttributeError:
            return http.NotFound()

        return self.execute(handler)




class ValidationError(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self): # pragma: no cover
        return 'ValidationError: %s' % self.message


class ValidationErrors(Exception):
    '''A container for a dict of validation errors.'''
    def __init__(self, errors):
        self.errors = errors

    def __str__(self):
        return 'ValidationErrors: %s' % (
            '\n'.join('%s: %s' % (
                (key, map(str, value))
                for key, value in self.errors.items()
            )),
        )



from __future__ import unicode_literals

'''Add some missing HttpResponse sub-classes'''
from django.core.exceptions import SuspiciousOperation
from django.http import HttpResponse, Http404
from django.utils.encoding import iri_to_uri

try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

import json
import re

STATUS_CODES = (
    (100, 'Continue'),
    (101, 'Switching Protocols'),

    (200, 'OK'),
    (201, 'Created'),
    (202, 'Accepted'),
    #(203, Non-Authoritative Information),
    (204, 'No Content'),
    (205, 'Reset Content'),
    (206, 'Partial Content'),

    (300, 'Multiple Choices'),
    (301, 'Moved Permanently'),
    (302, 'Found'),
    (303, 'See Other'),
    (304, 'Not Modified'),
    (305, 'Use Proxy'),
    #306 Deprecated
    (307, 'Temporary Redirect'),
    (308, 'Permanent Redirect'),

    (400, 'Bad Request'),
    (401, 'Unauthorized'),
    (402, 'Payment Required'),
    (403, 'Forbidden'),
    (404, 'Not Found'),
    (405, 'Method Not Allowed'),
    (406, 'Not Acceptable'),
    (407, 'Proxy Authentication Required'),
    (408, 'Request Timeout'),
    (409, 'Conflict'),
    (410, 'Gone'),
    (411, 'Length Required'),
    (412, 'Precondition Failed'),
    (413, 'Request Entity Too Large'),
    (414, 'Request-URI Too Long'),
    (415, 'Unsupported Media Type'),
    (416, 'Requested Range Not Satisfiable'),
    (417, 'Expectation Failed'),

    (500, 'Internal Server Error'),
    (501, 'Not Implemented'),
    (502, 'Bad Gateway'),
    (503, 'Service Unavailable'),
    (504, 'Gateway Timeout'),
    (505, 'HTTP Version Not Supported'),
)

STATUS = OrderedDict(STATUS_CODES)
# Set constant-like properties for reverse lookup
for code, label in STATUS_CODES:
    setattr(STATUS, re.sub(r'\W', '_', label.upper()), code)

class BaseHttpResponse(HttpResponse, Exception):
    '''
    A sub-class of HttpResponse that is also an Exception, allowing us to
    raise/catch it.

    With thanks to schinkel's repose.
    '''

#
# Success Responses (2xx)
#

class HttpResponseSuccess(BaseHttpResponse):
    '''A base class for all 2xx responses, so we can issubclass test.'''

class OK(HttpResponseSuccess):
    status_code = STATUS.OK

class Created(HttpResponseSuccess):
    status_code = STATUS.CREATED

class Accepted(HttpResponseSuccess):
    status_code = STATUS.ACCEPTED

class NoContent(HttpResponseSuccess):
    status_code = STATUS.NO_CONTENT

class ResetContent(HttpResponseSuccess):
    status_code = STATUS.RESET_CONTENT

class PartialContent(HttpResponseSuccess):
    status_code = STATUS.PARTIAL_CONTENT

#
# Redirection Responses (3xx)
#

class HttpResponseRedirection(BaseHttpResponse):
    '''A base class for all 3xx responses.'''

class LocationHeaderMixin(object):
    '''Many 3xx responses require a Location header'''
    def __init__(self, location, *args, **kwargs):
        super(LocationHeaderMixin, self).__init__(*args, **kwargs)
        parsed = urlparse(location)
        if parsed.scheme and parsed.scheme not in self.allowed_schemes:
            raise SuspiciousOperation(
                "Unsafe redirect to URL with protocol '%s'" % parsed.scheme
            )
        self['Location'] = iri_to_uri(location)

    url = property(lambda self: self['Location'])


class MultipleChoices(HttpResponseRedirection):
    status_code = STATUS.MULTIPLE_CHOICES

class MovedPermanently(LocationHeaderMixin, HttpResponseRedirection):
    status_code = STATUS.MOVED_PERMANENTLY

class Found(LocationHeaderMixin, HttpResponseRedirection):
    status_code = STATUS.FOUND

class SeeOther(LocationHeaderMixin, HttpResponseRedirection):
    status_code = STATUS.SEE_OTHER

class NotModified(HttpResponseRedirection):
    status_code = STATUS.NOT_MODIFIED

class UseProxy(LocationHeaderMixin, HttpResponseRedirection):
    status_code = STATUS.USE_PROXY

class TemporaryRedirect(HttpResponseRedirection):
    status_code = STATUS.TEMPORARY_REDIRECT

class PermanentRedirect(HttpResponseRedirection):
    status_code = STATUS.PERMANENT_REDIRECT

#
# Client Error Responses (4xx)
#

class HttpResponseError(BaseHttpResponse):
    '''A base class for all 4xx responses.'''

class BadRequest(HttpResponseError):
    status_code = STATUS.BAD_REQUEST

# XXX Auth-Realm ?
class Unauthorized(HttpResponseError):
    status_code = STATUS.UNAUTHORIZED

class PaymentRequired(HttpResponseError):
    status_code = STATUS.PAYMENT_REQUIRED

class Forbidden(HttpResponseError):
    status_code = STATUS.FORBIDDEN

class NotFound(HttpResponseError):
    status_code = STATUS.NOT_FOUND

class MethodNotAllowed(HttpResponseError):
    def __init__(self, permitted_methods, *args, **kwargs):
        super(MethodNotAllowed, self).__init__(*args, **kwargs)
        self['Allow'] = ', '.join(permitted_methods)

    status_code = STATUS.METHOD_NOT_ALLOWED

class NotAcceptable(HttpResponseError):
    status_code = STATUS.NOT_ACCEPTABLE

class ProxyAuthenticationRequired(HttpResponseError):
    status_code = STATUS.PROXY_AUTHENTICATION_REQUIRED

class RequestTimeout(HttpResponseError):
    status_code = STATUS.REQUEST_TIMEOUT

class Conflict(HttpResponseError):
    status_code = STATUS.CONFLICT

class Gone(HttpResponseError):
    status_code = STATUS.GONE

class LengthRequired(HttpResponseError):
    status_code = STATUS.LENGTH_REQUIRED

class PreconditionFailed(HttpResponseError):
    status_code = STATUS.PRECONDITION_FAILED

class RequestEntityTooLarge(HttpResponseError):
    status_code = STATUS.REQUEST_ENTITY_TOO_LARGE

class RequestURITooLong(HttpResponseError):
    status_code = STATUS.REQUEST_URI_TOO_LONG

class UnsupportedMediaType(HttpResponseError):
    status_code = STATUS.UNSUPPORTED_MEDIA_TYPE

class RequestedRangeNotSatisfiable(HttpResponseError):
    status_code = STATUS.REQUESTED_RANGE_NOT_SATISFIABLE

class ExpectationFailed(HttpResponseError):
    status_code = STATUS.EXPECTATION_FAILED

#
# Server Error (5xx)
#

class HttpResponseServerError(BaseHttpResponse):
    '''A base class for 5xx responses.'''

class InternalServerError(HttpResponseServerError):
    status_code = STATUS.INTERNAL_SERVER_ERROR

class NotImplemented(HttpResponseServerError):
    status_code = STATUS.NOT_IMPLEMENTED

class BadGateway(HttpResponseServerError):
    status_code = STATUS.BAD_GATEWAY

class ServiceUnavailable(HttpResponseServerError):
    status_code = STATUS.SERVICE_UNAVAILABLE

class GatewayTimeout(HttpResponseServerError):
    status_code = STATUS.GATEWAY_TIMEOUT

class HttpVersiontNotSupported(HttpResponseServerError):
    status_code = STATUS.HTTP_VERSION_NOT_SUPPORTED

#
# General Helpers
#

class JsonResponse(HttpResponse):
    '''Handy shortcut for dumping JSON data'''
    def __init__(self, content, *args, **kwargs):
        kwargs.setdefault('content_type', 'application/json')
        super(JsonResponse, self).__init__(json.dumps(content), *args, **kwargs)



from __future__ import unicode_literals

from six import with_metaclass

class DeclarativeMetaclass(type):
    '''
    Mangle a class so any properties defined on it are stashed in _defaults
    '''
    def __new__(mcs, name, bases, attrs):
        defaults = {
            key: attrs.pop(key)
            for key in list(attrs.keys())
            if not key.startswith('_')
        }
        new_class = super(DeclarativeMetaclass, mcs).__new__(mcs, name, bases,
            attrs)
        new_class._defaults = defaults
        return new_class


class Meta(with_metaclass(DeclarativeMetaclass, object)):
    '''Generic container for Meta classes'''

    def __init__(self, meta):
        ''' Copy value values onto ourself '''
        for key in dir(meta):
            if key.startswith('_'):
                continue
            if key not in self._defaults:
                raise ValueError('Property "%s" not supported by %s' % (key, self.__class__.__name__))
        for key, value in self._defaults.items():
            setattr(self, key, getattr(meta, key, value))



from __future__ import unicode_literals

from .publisher import Publisher, SimplePatternsMixin
from .models import ModelPublisher

from django.template.response import TemplateResponse

class ViewPublisherMixin(object):
    '''
    A Publisher for non-API views.

    Replaces create_response with a wrapper around TemplateResponse.
    '''
    api_name = None

    def create_response(self, content, **response_kwargs):
        '''Return a response serialising the content'''
        template = response_kwargs.pop('template', None)
        if not template:
            template = self.get_template_names()
        return TemplateResponse(self.request, template, content,
            **response_kwargs
        )

class ViewPublisher(SimplePatternsMixin, Publisher):
    def get_template_name(self):
        return self.template_name % {
            'action': self.action,
            'method': self.method,
            'mode': self.mode,
        }

class ModelViewPublisher(SimplePatternsMixin, ModelPublisher):
    model = None

    def get_template_name(self):
        '''Return the template name to use for this view.'''
        return '%s/%s_%s.html' % (
            self.api_name or self.model._meta.app_label,
            self.model._meta.model_name,
            self.action,
        )



from __future__ import unicode_literals

# Authentication and Authorisation
from functools import wraps
from . import http

def permit(test_func):
    '''Decorate a handler to control access'''
    def decorator(view_func):
        @wraps(view_func)
        def _wrapped_view(self, *args, **kwargs):
            if test_func(self, *args, **kwargs):
                return view_func(self, *args, **kwargs)
            return http.Forbidden()
        return _wrapped_view
    return decorator

permit_logged_in = permit(
    lambda self, *args, **kwargs: self.request.user.is_authenticated()
)

permit_staff = permit(
    lambda self, *args, **kwargs: self.request.user.is_staff
)


def permit_groups(*groups):
    def in_groups(request, *args):
        return request.user.groups.filter(name__in=args).exists()
    return permit(
        lambda self, *args, **kwargs: in_groups(self.request, *groups)
    )




from django.http import StreamingHttpResponse
from django.utils.encoding import force_text

from nap.models import modelserialiser_factory
from .simplecsv import Writer


class ExportCsv(object):
    '''
    A factory class for Admin Actions to export a model as CSV.

    actions = [ ExportCsv(myserialiser, 'Export as CSV'), ]

    Optionally, you can pass kwargs to be passed to modelserialiser_factory to
    create a ModelSerialiser.
    '''
    def __init__(self, serialiser=None, label=None, **opts):
        self.serialiser = serialiser
        self.opts = opts
        if label:
            self.short_description = label
        self.__name__ = label or 'ExportCsv'

    def __call__(self, admin, request, queryset):
        if self.serialiser is None:
            ser_class = modelserialiser_factory(
                '%sSerialiser' % admin.__class__.__name__,
                admin.model,
                **self.opts
            )
        else:
            ser_class = self.serialiser

        select_related = self.opts.get('select_related', None)
        if select_related:
            queryset = queryset.select_related(*select_related)
        prefetch_related = self.opts.get('prefetch_related', None)
        if prefetch_related:
            queryset = queryset.prefetch_related(*prefetch_related)

        def inner(ser):
            csv = Writer(fields=self.opts.get('fields', ser._fields.keys()))
            yield csv.write_headers()
            for obj in queryset:
                data = { 
                    key: force_text(val)
                    for key, val in ser.object_deflate(obj).items()
                }
                yield csv.write_dict(data)

        response = StreamingHttpResponse(inner(ser_class()), content_type='text/csv')
        filename = self.opts.get('filename', 'export_{classname}.csv')
        if callable(filename):
            filename = filename(admin)
        else:
            filename = filename.format(
                classname=admin.__class__.__name__,
                model=admin.model._meta.module_name,
                app_label=admin.model._meta.app_label,
            )
        response['Content-Disposition'] = 'attachment; filename=%s' % filename

        return response



from __future__ import unicode_literals

from . import fields
from .meta import Meta
from .publisher import Publisher
from .serialiser import MetaSerialiser, Serialiser

from six import with_metaclass

from django.db.models import Manager
from django.db.models.fields import NOT_PROVIDED
from django.shortcuts import get_object_or_404

FIELD_MAP = {}
# Auto-construct the field map
for f in dir(fields):
    cls = getattr(fields, f)
    try:
        if issubclass(cls, fields.Field):
            FIELD_MAP[f] = cls
    except TypeError:
        pass


class ModelMeta(Meta):
    model = None
    # Fields to include from the Model
    fields = []
    field_types = {}
    # Fields to exclude from the Model
    exclude = []
    # Fields from the model to flag as read-only
    read_only_fields = []

    # When restoring an object, which fields do we treat as m2m?
    related_fields = []
    # When we try to retrieve an instance, which fields do we use?
    key_fields = ('id',)
    # When restoring an object, which attrs do we not use
    ignored_fields = ()
    # When calling get_or_create, what do we pass as defaults?
    defaults = {}
    # Are there any deserialised fields we must pass to get_or_create defaults?
    core_fields = ()


class MetaModelSerialiser(MetaSerialiser):
    meta_class = ModelMeta

    def __new__(mcs, name, bases, attrs):

        new_class = super(MetaModelSerialiser, mcs).__new__(mcs, name, bases, attrs)

        include = getattr(new_class._meta, 'fields', [])
        exclude = getattr(new_class._meta, 'exclude', [])
        read_only = getattr(new_class._meta, 'read_only_fields', [])

        current_fields = new_class._fields.keys()

        model_fields = {}
        model = new_class._meta.model
        if model is not None:
            for field in model._meta.fields:
                # If we've got one, skip...
                if field.name in current_fields:
                    continue

                # If we have a whitelist, and it's not in it, skip
                if include and not field.name in include:
                    continue

                # If it's blacklisted, skip
                if field.name in exclude:
                    continue

                kwargs = {
                    'readonly': field.name in read_only,
                    'null': field.null,
                }
                if not field.default is NOT_PROVIDED:
                    kwargs['default'] = field.default

                try:
                    field_class = new_class._meta.field_types[field.name]
                except KeyError:
                    field_class = FIELD_MAP.get(field.__class__.__name__, fields.Field)
                model_fields[field.name ] = field_class(**kwargs)

        new_class._fields.update(model_fields)

        return new_class


class ModelSerialiser(with_metaclass(MetaModelSerialiser, Serialiser)):

    def restore_object(self, obj, instance, **kwargs):
        if instance:
            for key, val in obj.items():
                setattr(instance, key, val)
        else:
            instance = self._meta.model(**obj)
        if kwargs.get('commit', True):
            instance.save()
        return instance


class ModelReadSerialiser(ModelSerialiser):
    '''
    A Serialiser that will only Read, not Create, an instance on inflate.

    Specify only the fields you intend to filter by.
    '''

    def restore_object(self, objdata, instance, **kwargs):
        try:
            return self._meta.model.objects.get(**objdata)
        except self._meta.model.DoesNotExist:
            raise fields.ValidationError('%s with values %r does not exist.' % (
                self._meta.model.__name__,
                objdata,
            ))


class ModelCreateUpdateSerialiser(ModelSerialiser):

    '''
    A ModelSerialiser with the ability to create/update instances from data.
    '''

    def restore_object(self, objdata, instance, **kwargs):
        related = {
            key: objdata.pop(key)
            for key in self._meta.related_fields
        }

        if instance is None:
            lookup = {
                field: objdata[field]
                for field in self._meta.key_fields
            }
            defaults = dict(self._meta.defaults)
            for key in self._meta.core_fields:
                defaults[key] = objdata[key]
            instance, _ = self._meta.model.objects.get_or_create(
                defaults=defaults,
                **lookup
            )

        for key, value in objdata.items():
            if key in self._meta.ignored_fields:
                continue
            setattr(instance, key, value)

        instance.save()

        for key, value in related.items():
            getattr(instance, key).add(*value)

        return instance


class ModelPublisher(Publisher):
    '''A Publisher with useful methods to publish Models'''

    @property
    def model(self):
        '''By default, we try to get the model from our serialiser'''
        # XXX Should this call get_serialiser?
        return self.serialiser._meta.model

    # Auto-build serialiser from model class?

    def get_object_list(self):
        return self.model.objects.all()

    def get_object(self, object_id):
        return get_object_or_404(self.get_object_list(), pk=object_id)


def modelserialiser_factory(name, model, **kwargs):
    attrs = {
        'model': model,
    }
    for key in ['fields', 'exclude', 'read_only']:
        try:
            attrs[key] = kwargs[key]
        except KeyError:
            pass

    meta = type(str('Meta'), (object,), attrs)
    return type(ModelSerialiser)(str(name), (ModelSerialiser,), {'Meta': meta})


class ModelSerialiserField(fields.SerialiserField):
    def __init__(self, *args, **kwargs):
        if not 'serialiser' in kwargs:
            model = kwargs.pop('model')
            kwargs['serialiser'] = modelserialiser_factory(model.__name__ + 'Serialiser', model, **kwargs)()
        super(ModelSerialiserField, self).__init__(*args, **kwargs)


class ModelManySerialiserField(fields.ManySerialiserField):
    def __init__(self, *args, **kwargs):
        if not 'serialiser' in kwargs:
            model = kwargs.pop('model')
            kwargs['serialiser'] = modelserialiser_factory(model.__name__ + 'Serialiser', model, **kwargs)()
        super(ModelManySerialiserField, self).__init__(*args, **kwargs)

    def reduce(self, value, **kwargs):
        if isinstance(value, Manager):
            value = value.all()
        return super(ModelManySerialiserField, self).reduce(value, **kwargs)




from django.test import TestCase

from nap import fields
from nap.exceptions import ValidationError

class Mock(object):
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

class FieldTestCase(TestCase):
    '''
    Field cycles:
    deflate: digattr -> reduce -> data[name]
    inflate: data[name] -> restore -> dest[name]
    '''

    def test_000_field(self):
        data = {}
        field = fields.Field()

        field.deflate('value', Mock(value='str'), data)
        self.assertEqual('str', data['value'])

        dest = {}

        field.inflate('value', data, dest)
        self.assertEqual(dest['value'], data['value'])

    def test_000_field_none(self):
        '''None is treated specially.'''
        data = {}
        field = fields.Field()

        field.deflate('value', Mock(value=None), data)
        self.assertTrue(data['value'] is None)

        dest = {}

        field.inflate('value', data, dest)
        self.assertEqual(data['value'], dest['value'])

    def test_000_field_default(self):
        '''
        With no default set, we can't deflate a field with no value.
        '''
        data = {}
        field = fields.Field()

        with self.assertRaises(AttributeError):
            field.deflate('value', Mock(), data)

    def test_000_readonly(self):
        data = {'value': 1}
        dest = {}
        field = fields.Field(readonly=True)
        field.inflate('value', data, dest)
        self.assertNotIn('value', dest)

    def test_000_nodefault(self):
        data = {}
        dest = {}
        field = fields.Field()
        field.inflate('value', data, dest)
        self.assertNotIn('value', dest)

    def test_000_default(self):
        data = {}
        dest = {}
        field = fields.Field(default=1)

        field.inflate('value', data, dest)
        self.assertEqual(dest['value'], 1)

    def test_000_notnull(self):
        data = {'value': None}
        dest = {}
        field = fields.Field(null=False)

        with self.assertRaises(ValidationError):
            field.inflate('value', data, dest)

    def test_001_boolean(self):
        data = {}
        field = fields.BooleanField()

        field.deflate('value', Mock(value=True), data)
        self.assertTrue(data['value'] is True)

        dest = {}

        field.inflate('value', data, dest)
        self.assertTrue(dest['value'] is True)

    def test_002_integer(self):
        data = {}
        field = fields.IntegerField()

        field.deflate('value', Mock(value=7), data)
        self.assertEqual(data['value'], 7)

        dest = {}

        field.inflate('value', data, dest)
        self.assertEqual(dest['value'], data['value'])

    def test_003_decimal(self):
        from decimal import Decimal
        data = {}
        field = fields.DecimalField()

        field.deflate('value', Mock(value=Decimal('1.05')), data)
        # JS has no Decimal type, only float
        self.assertEqual(data['value'], 1.05)

        dest = {}

        field.inflate('value', data, dest)
        self.assertEqual(dest['value'], data['value'])

    def test_004_datetime(self):
        from datetime import datetime
        data = {}
        field = fields.DateTimeField()

        when = datetime(2010, 11, 5, 12, 7, 19)
        field.deflate('value', Mock(value=when), data)
        self.assertEqual(data['value'], '2010-11-05 12:07:19')

        dest = {}

        field.inflate('value', data, dest)
        self.assertEqual(dest['value'], when)

    def test_005_date(self):
        from datetime import date
        data = {}
        field = fields.DateField()

        when = date(2010, 11, 5)
        field.deflate('value', Mock(value=when), data)
        self.assertEqual(data['value'], '2010-11-05')

        dest = {}

        field.inflate('value', data, dest)
        self.assertEqual(dest['value'], when)

    def test_006_time(self):
        from datetime import time
        data = {}
        field = fields.TimeField()

        when = time(12, 7, 19)
        field.deflate('value', Mock(value=when), data)
        self.assertEqual(data['value'], '12:07:19')

        dest = {}

        field.inflate('value', data, dest)
        self.assertEqual(dest['value'], when)

    def test_007_serialiser(self):
        from nap.serialiser import Serialiser
        class SimpleSerialiser(Serialiser):
            a = fields.Field()
            b = fields.Field()
            c = fields.Field()

        data = {}
        field = fields.SerialiserField(serialiser=SimpleSerialiser())

        value = Mock(value=Mock(a=1, b='two', c=3.0))
        field.deflate('value', value, data)
        self.assertEqual(data['value']['a'], 1)
        self.assertEqual(data['value']['b'], 'two')
        self.assertEqual(data['value']['c'], 3.0)

